use std::{
    collections::VecDeque,
    fmt::Display,
    num::NonZeroUsize,
    ops::{Deref, DerefMut},
};

use bitvec::prelude::*;
use log::{error, warn};
use rand::prelude::*;

use super::{wire::Request, Coord, Direction};

#[derive(Debug, PartialEq, Eq, Clone)]
enum SmallBitBox {
    Stack {
        storage: BitArr!(for Self::STACK_BITS),
        len: NonZeroUsize,
    },
    Heap(BitBox),
}

impl SmallBitBox {
    const STACK_BITS: usize = usize::BITS as usize * 3;

    fn new(initial: bool, len: usize) -> Self {
        if let len @ 1..Self::STACK_BITS = len {
            let Some(len) = NonZeroUsize::new(len) else {
                unreachable!()
            };
            let mut storage = BitArray::ZERO;
            storage.fill(initial);
            Self::Stack { storage, len }
        } else {
            Self::Heap(bitbox![u8::from(initial); len])
        }
    }
}

impl Deref for SmallBitBox {
    type Target = BitSlice;

    fn deref(&self) -> &Self::Target {
        match self {
            Self::Stack { storage, len } => &storage[..len.get()],
            Self::Heap(bit_box) => bit_box,
        }
    }
}

impl DerefMut for SmallBitBox {
    fn deref_mut(&mut self) -> &mut Self::Target {
        match self {
            Self::Stack { storage, len } => &mut storage[..len.get()],
            Self::Heap(bit_box) => bit_box,
        }
    }
}

#[derive(Debug, PartialEq, Eq, Clone)]
pub struct Board {
    width: u8,
    height: u8,
    hazard_damage: u8,
    food_spawn_chance: u8,
    min_food: u16,
    turn: u32,
    food: SmallBitBox,
    hazard: SmallBitBox,
    free: SmallBitBox,
    snakes: Vec<Snake>,
    constrictor: bool,
}

impl From<&Request> for Board {
    fn from(value: &Request) -> Self {
        let width = value.board.width;
        let height = value.board.height;
        let fields = usize::from(width) * usize::from(height);
        let mut board = Self {
            width,
            height,
            hazard_damage: value.game.ruleset.settings.hazard_damage_per_turn,
            food_spawn_chance: value.game.ruleset.settings.food_spawn_chance,
            min_food: value.game.ruleset.settings.minimum_food,
            turn: value.turn,
            food: SmallBitBox::new(false, fields),
            hazard: SmallBitBox::new(false, fields),
            free: SmallBitBox::new(true, fields),
            snakes: Vec::with_capacity(value.board.snakes.len()),
            constrictor: value.game.ruleset.name == "constrictor",
        };

        for &food in &value.board.food {
            let index = board.coord_to_linear(food);
            board.food.set(index, true);
        }

        for &hazard in &value.board.hazards {
            let index = board.coord_to_linear(hazard);
            board.hazard.set(index, true);
        }

        for (id, snake) in value.board.snakes.iter().enumerate() {
            for &tile in snake
                .body
                .iter()
                .take(snake.body.len() - usize::from(!board.constrictor))
            {
                let index = board.coord_to_linear(tile);
                board.free.set(index, false);
            }
            let snake = Snake {
                body: snake.body.iter().rev().copied().collect(),
                id: u8::try_from(id).unwrap_or(u8::MAX),
                health: snake.health,
            };
            board.snakes.push(snake);
        }

        board
    }
}

impl Display for Board {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(
            f,
            "{} {}x{} {}% ({}) {}dmg @ {}",
            if self.constrictor { "constrictor" } else { "" },
            self.width,
            self.height,
            self.food_spawn_chance,
            self.min_food,
            self.hazard_damage,
            self.turn
        )?;

        for y in (0..self.height).rev() {
            for x in 0..self.width {
                let tile = Coord { x, y };
                if self.snakes.iter().any(|snake| snake.head() == tile) {
                    write!(f, "H")?;
                    continue;
                }
                if self.snakes.iter().any(|snake| snake.tail() == tile) {
                    write!(f, "T")?;
                    continue;
                }

                let index = self.coord_to_linear(tile);
                if !self.free[index] {
                    write!(f, "S")?;
                    continue;
                }
                if self.food[index] {
                    write!(f, "f")?;
                    continue;
                }
                if self.hazard[index] {
                    write!(f, "h")?;
                    continue;
                }

                write!(f, ".")?;
            }
            writeln!(f)?;
        }

        Ok(())
    }
}

impl Board {
    #[must_use]
    pub const fn turn(&self) -> u32 {
        self.turn
    }

    #[must_use]
    pub fn num_snakes(&self) -> usize {
        self.snakes.len()
    }

    #[must_use]
    pub fn is_food(&self, tile: Coord) -> bool {
        let index = self.coord_to_linear(tile);
        self.food[index]
    }

    #[must_use]
    pub fn is_hazard(&self, tile: Coord) -> bool {
        let index = self.coord_to_linear(tile);
        self.hazard[index]
    }

    #[must_use]
    pub fn is_free(&self, tile: Coord) -> bool {
        if !(tile.x < self.width && tile.y < self.height) {
            return false;
        }
        let index = self.coord_to_linear(tile);
        self.free[index]
    }

    pub fn valid_actions(&self, id: u8) -> impl Iterator<Item = Direction> + use<'_> {
        let head = self
            .snakes
            .binary_search_by_key(&id, |snake| snake.id)
            .ok()
            .map(|index| self.snakes[index].head());
        if head.is_none() {
            warn!(
                "Asked for an action for a snake that doesn't exist: {id} not in {:?}",
                self.snakes
            );
        }
        enum_iterator::all::<Direction>()
            .filter_map(move |direction| {
                head.and_then(|head| head.apply(direction))
                    .map(|tile| (direction, tile))
            })
            .filter(|(_, tile)| tile.x < self.width && tile.y < self.height)
            .filter(|(_, tile)| self.is_free(*tile))
            .map(|(direction, _)| direction)
    }

    pub fn simulate_random<T>(&mut self, stop: impl Fn(&Self) -> Option<T>) -> T {
        loop {
            if let Some(score) = stop(self) {
                break score;
            }
            self.next_turn(&[]);
        }
    }

    pub fn next_turn(&mut self, actions: &[(u8, Direction)]) {
        self.move_standard(actions);
        self.starvation_standard();
        self.hazard_damage_standard();
        self.feed_snakes_standard();
        self.eliminate_snake_standard();
        self.update_free_map();
        self.spawn_food();
        self.turn += 1;
    }

    fn move_standard(&mut self, actions: &[(u8, Direction)]) {
        for i in 0..self.snakes.len() {
            let snake = &self.snakes[i];
            let action = actions.iter().find(|(id, _)| *id == snake.id).map_or_else(
                || {
                    self.valid_actions(snake.id)
                        .choose(&mut thread_rng())
                        .unwrap_or(Direction::Up)
                },
                |(_, action)| *action,
            );
            let new_head = snake.head().wrapping_apply(action);
            self.snakes[i].advance(new_head);
        }
    }

    fn starvation_standard(&mut self) {
        for snake in &mut self.snakes {
            snake.health = snake.health.saturating_sub(1);
        }
    }

    fn hazard_damage_standard(&mut self) {
        let mut i = 0;
        while i < self.snakes.len() {
            let head = self.snakes[i].head();
            if self.is_in_bounds(head) {
                let head_index = self.coord_to_linear(head);
                if self.hazard[head_index] && !self.food[head_index] {
                    let health = &mut self.snakes[i].health;
                    *health = health.saturating_sub(1);
                    if *health == 0 {
                        let snake = self.snakes.remove(i);
                        for tile in snake.body {
                            let index = self.coord_to_linear(tile);
                            self.free.set(index, true);
                        }
                        continue;
                    }
                }
            }
            i += 1;
        }
    }

    fn feed_snakes_standard(&mut self) {
        if self.constrictor {
            for snake in &mut self.snakes {
                snake.feed();
            }
        } else {
            let mut eaten_food = vec![];
            for i in 0..self.snakes.len() {
                let head = self.snakes[i].head();
                if self.is_in_bounds(head) {
                    let head_index = self.coord_to_linear(head);
                    if self.food[head_index] {
                        eaten_food.push(head_index);
                        self.snakes[i].feed();
                    }
                }
            }
            for food_index in eaten_food {
                self.food.set(food_index, false);
            }
        }
    }

    fn eliminate_snake_standard(&mut self) {
        // eliminate out of health and out of bounds
        let mut i = 0;
        while i < self.snakes.len() {
            let snake = &self.snakes[i];
            if snake.health == 0 || !self.is_in_bounds(snake.head()) {
                let snake = self.snakes.remove(i);
                for tile in snake.body {
                    if self.is_in_bounds(tile) {
                        let index = self.coord_to_linear(tile);
                        self.free.set(index, true);
                    }
                }
                continue;
            }
            i += 1;
        }

        // look for collisions
        let mut collisions = vec![];
        for snake in &self.snakes {
            let head = snake.head();
            let head_index = self.coord_to_linear(head);
            if !self.free[head_index] {
                collisions.push(snake.id);
                continue;
            }
            for snake2 in &self.snakes {
                if snake.id != snake2.id
                    && snake.head() == snake2.head()
                    && snake.body.len() <= snake2.body.len()
                {
                    collisions.push(snake.id);
                    break;
                }
            }
        }

        // apply collisions
        let mut i = 0;
        while i < self.snakes.len() {
            if collisions.contains(&self.snakes[i].id) {
                let snake = self.snakes.remove(i);
                for tile in snake.body {
                    let index = self.coord_to_linear(tile);
                    self.free.set(index, true);
                }
                continue;
            }
            i += 1;
        }
    }

    fn update_free_map(&mut self) {
        // free tails
        for snake in &self.snakes {
            let tail = snake.tail();
            let pre_tail = snake.body[snake.body.len() - 2];
            if tail != pre_tail {
                let tail_index = self.coord_to_linear(tail);
                self.free.set(tail_index, true);
            }
        }
        // block heads
        for snake in &self.snakes {
            let head = snake.head();
            let head_index = self.coord_to_linear(head);
            self.free.set(head_index, false);
        }
    }

    fn spawn_food(&mut self) {
        let num_food = self.food.count_ones();
        let needed_food = if num_food < usize::from(self.min_food) {
            usize::from(self.min_food) - num_food
        } else {
            usize::from(
                self.food_spawn_chance > 0
                    && thread_rng().gen_range(0..100) < self.food_spawn_chance,
            )
        };

        let food_spots = self
            .free
            .iter()
            .enumerate()
            .filter_map(|(i, free)| free.then_some(i))
            .choose_multiple(&mut thread_rng(), needed_food);
        for index in food_spots {
            self.food.set(index, true);
        }
    }

    fn coord_to_linear(&self, coord: Coord) -> usize {
        usize::from(coord.x) + usize::from(coord.y) * usize::from(self.width)
    }

    const fn is_in_bounds(&self, coord: Coord) -> bool {
        coord.x < self.width && coord.y < self.height
    }
}

#[derive(Debug, PartialEq, Eq, Clone)]
struct Snake {
    id: u8,
    health: u8,
    body: VecDeque<Coord>,
}

impl Snake {
    pub fn head(&self) -> Coord {
        self.body.back().copied().unwrap_or_else(|| {
            error!("Snake without a head: {self:?}");
            Coord { x: 0, y: 0 }
        })
    }

    pub fn tail(&self) -> Coord {
        self.body.front().copied().unwrap_or_else(|| {
            error!("Snake without a tail: {self:?}");
            Coord { x: 0, y: 0 }
        })
    }

    pub fn advance(&mut self, head: Coord) {
        self.body.push_back(head);
        self.body.pop_front();
        self.health = 100;
    }

    pub fn feed(&mut self) {
        if let Some(tail) = self.body.front() {
            self.body.push_front(*tail);
        }
    }
}
